@@ -1,6 +1,5 @@
 // SPDX-License-Identifier: GPL-3.0-or-later
 
-use std::sync::Arc;
 use std::sync::Mutex;
 
 #[cfg(feature = "latest-gtk")]
@@ -29,10 +28,11 @@ pub use widgets::*;
 ///
 /// Side effects are implemented in middlewares (See [Middleware](crate::Middleware)).
 pub struct Store<S: std::fmt::Debug + Clone + Default + PartialEq + Eq + 'static> {
-    state: Arc<Mutex<S>>,
-    reducer: Arc<Mutex<Box<dyn Fn(&Action, &mut S) + 'static>>>,
-    selectors: Arc<Mutex<Vec<Selector<S>>>>,
-    middlewares: Arc<Mutex<Arc<Vec<Box<dyn Middleware<S>>>>>>,
+    state: S,
+    middlewares: Vec<Box<dyn Middleware<S>>>,
+    reducer: Box<dyn Fn(&Action, &mut S) + 'static>,
+    selectors: Vec<Selector<S>>,
+    reduce_lock: Mutex<u8>,
 }
 
 impl<S: std::fmt::Debug + Clone + Default + PartialEq + Eq + 'static> std::fmt::Debug for Store<S> {
@@ -56,31 +56,25 @@ impl<S: std::fmt::Debug + Clone + Default + PartialEq + Eq> Store<S> {
         middlewares: Vec<Box<dyn Middleware<S>>>,
     ) -> Self {
         Store {
-            state: Arc::new(Mutex::new(default_state)),
-            reducer: Arc::new(Mutex::new(Box::new(reducer))),
-            selectors: Arc::new(Mutex::new(Default::default())),
-            middlewares: Arc::new(Mutex::new(Arc::new(middlewares))),
+            state: default_state,
+            middlewares,
+            reducer: Box::new(reducer),
+            selectors: Default::default(),
+            reduce_lock: Default::default(),
         }
     }
 
     fn select_internal(
-        &self,
+        &mut self,
         name: &str,
         selector: impl Fn(&S) -> S + 'static,
-        callback: impl Fn(S) + 'static,
+        callback: impl Fn(&S) + 'static,
         select_full_state: bool,
     ) {
-        let state = self.state.lock().unwrap();
-        let s = state.clone();
-        drop(state);
-        callback(s);
-
-        let mut selectors = self.selectors.lock().unwrap();
-        let state = self.state.lock().unwrap();
-
-        let last_selector_state = selector(&state);
+        callback(&self.state);
+        let last_selector_state = selector(&self.state);
         trace!("Adding selector for 'state[{}]'.", name);
-        selectors.push(Selector::new(
+        self.selectors.push(Selector::new(
             name,
             select_full_state,
             selector,
@@ -91,10 +85,10 @@ impl<S: std::fmt::Debug + Clone + Default + PartialEq + Eq> Store<S> {
 
     /// Select a slice of the state where everything else is Default::default().
     pub fn select(
-        &self,
+        &mut self,
         name: &str,
         selector: impl Fn(&S) -> S + 'static,
-        callback: impl Fn(S) + 'static,
+        callback: impl Fn(&S) + 'static,
     ) {
         print_perf!(
             self.select_internal(name, selector, callback, false);
@@ -104,10 +98,10 @@ impl<S: std::fmt::Debug + Clone + Default + PartialEq + Eq> Store<S> {
 
     /// Select a slice of the state with receiving the whole state.
     pub fn select_full(
-        &self,
+        &mut self,
         name: &str,
         selector: impl Fn(&S) -> S + 'static,
-        callback: impl Fn(S) + 'static,
+        callback: impl Fn(&S) + 'static,
     ) {
         print_perf!(
             self.select_internal(name, selector, callback, true);
@@ -129,85 +123,76 @@ impl<S: std::fmt::Debug + Clone + Default + PartialEq + Eq> Store<S> {
 
     /// Dispatch the given arguments as an [Action](crate::Action).
     #[cfg(feature = "glib")]
-    pub fn dispatch(&self, action: String, argument: Option<glib::Variant>) {
-        let m = format!("Dispatch action {:?} with argument {:?}", action, argument);
-        print_perf!(
+    pub fn dispatch(&mut self, action: String, argument: Option<glib::Variant>) {
+        let lock = self.reduce_lock.try_lock();
+        if lock.is_err() {
+            error!("Can not dispatch '{}' during reduce", action);
+            return;
+        }
+        drop(lock);
+
+        let message = format!("Dispatch action {:?} with argument {:?}", action, argument);
+        print_perf! (
             {
                 let action = Action::new(action, argument);
+                // let state = &mut self.state1;
 
-                trace!("Locking store...");
-                let mut state_lock = self.state.lock().unwrap();
-                let reducer = self.reducer.lock().unwrap();
-                let mut selectors = self.selectors.lock().unwrap();
-                let middlewares_lock = self.middlewares.lock().unwrap();
-                let middlewares = middlewares_lock.clone();
-                drop(middlewares_lock);
-
-                let state = &mut *state_lock;
                 trace!(
                     "Reduce action {:?} for state {:?}.",
                     action.name().to_string(),
-                    state
+                    &self.state
                 );
 
                 print_perf!(
-                    for middleware in middlewares.as_ref() {
-                        middleware.pre_reduce(&action, &state);
+                    for middleware in &self.middlewares {
+                        middleware.pre_reduce(&action, &mut self.state);
                     };
-                    format!("Call {} middlewares pre_reduce", middlewares.len())
+                    format!("Call {} middlewares pre_reduce", self.middlewares.len())
                 );
 
-                print_perf!(
-                    reducer(&action, state);
-                    format!("Reduce action {:?}", action.name())
-                );
-                trace!("Reduced to: {:?}", state);
-
-                let mut selectors_to_call = Vec::new();
-                print_perf!(
-                    {
-                        for selector in selectors.iter_mut() {
-                            let sel = &selector.selector;
-                            let selected_state = sel(&state);
-                            if selector.last_state != selected_state {
-                                selector.last_state = selected_state.clone();
-                                let callb = &selector.callback;
-
-                                if selector.full {
-                                    selectors_to_call.push((selector.name.clone(), callb.clone(), state.clone()));
-                                } else {
-                                    selectors_to_call.push((selector.name.clone(), callb.clone(), selected_state));
+                // reducing phase
+                {
+                    let lock = self.reduce_lock.try_lock();
+                    if lock.is_err() {
+                        error!("Can not dispatch '{}' during reduce", action.name());
+                        return;
+                    }
+                    let reducer = &self.reducer;
+                    print_perf!(
+                        reducer(&action, &mut self.state);
+                        format!("Reduce action {:?}", action.name())
+                    );
+                    trace!("Reduced to: {:?}", &self.state);
+                    print_perf!(
+                        {
+                            for selector in self.selectors.iter_mut() {
+                                let sel = &selector.selector;
+                                let selected_state = sel(&self.state);
+                                if selector.last_state != selected_state {
+                                    let callb = &selector.callback;
+                                    if selector.full {
+                                        callb(&self.state);
+                                    } else {
+                                        callb(&selected_state);
+                                    }
+                                    selector.last_state = selected_state;
                                 }
                             }
-                        }
-                    };
-                    format!("Determine n of {} selectors for action {:?}", selectors.len(), action.name())
-                );
-
-                let state_copy_for_middleware = state_lock.clone();
-                drop(state_lock);
-                drop(reducer);
-                drop(selectors);
-                trace!("Freed store.");
-
-                let selectors_to_call_len = selectors_to_call.len();
-                print_perf!(
-                    for (name, callback, state) in selectors_to_call {
-                        trace!("Run selector 'state[{}]'.", &name);
-                        callback(state)
-                    };
-                    format!("Call {} selectors", selectors_to_call_len)
-                );
+                        };
+                        format!("Determine n of {} selectors for action {:?}", self.selectors.len(), action.name())
+                    );
+                    drop(lock);
+                }
 
                 print_perf!(
-                    for middleware in middlewares.as_ref() {
-                        middleware.post_reduce(&action, &state_copy_for_middleware);
+                    for middleware in &self.middlewares {
+                        middleware.post_reduce(&action, &self.state);
                     };
-                    format!("Call {} middlewares post_reduce", middlewares.len())
+                    format!("Call {} middlewares post_reduce", self.middlewares.len())
                 );
             };
-            m
-        )
+            message
+        );
     }
 
     /// Delegate GTK actions as gstore::Actions to this store.
@@ -225,7 +210,7 @@ impl<S: std::fmt::Debug + Clone + Default + PartialEq + Eq> Store<S> {
     /// Delegate GTK actions as gstore::Actions to this store.
     #[cfg(feature = "latest-gtk")]
     pub fn delegate(
-        &'static self,
+        &'static mut self,
     ) -> glib::Sender<(gdk4::gio::SimpleAction, Option<glib::Variant>)> {
         let (sender, receiver) = glib::MainContext::channel::<(
             gdk4::gio::SimpleAction,
@@ -245,7 +230,7 @@ struct Selector<S: std::fmt::Debug + Clone + Default + 'static> {
     full: bool,
     selector: Box<dyn Fn(&S) -> S + 'static>,
     last_state: S,
-    callback: Arc<dyn Fn(S) + 'static>,
+    callback: Box<dyn Fn(&S) + 'static>,
 }
 
 impl<S: std::fmt::Debug + Clone + Default + 'static> Selector<S> {
@@ -254,14 +239,14 @@ impl<S: std::fmt::Debug + Clone + Default + 'static> Selector<S> {
         full: bool,
         selector: impl Fn(&S) -> S + 'static,
         last_state: S,
-        callback: impl Fn(S) + 'static,
+        callback: impl Fn(&S) + 'static,
     ) -> Self {
         Selector {
             name: name.to_string(),
             full,
             selector: Box::new(selector),
             last_state,
-            callback: Arc::new(callback),
+            callback: Box::new(callback),
         }
     }
 }
@@ -378,13 +363,13 @@ macro_rules! store {
         ) {
             unsafe {
                 STORE = Some(Store::new(default_state, reducer, middlewares));
-                STORE.as_ref().unwrap().dispatch(gstore::INIT.into(), None);
+                STORE.as_mut().unwrap().dispatch(gstore::INIT.into(), None);
             }
         }
 
         /// Get a static reference to the store
-        pub fn store() -> &'static Store {
-            unsafe { STORE.as_ref().expect("Store is not initialized!") }
+        pub fn store() -> &'static mut Store {
+            unsafe { STORE.as_mut().expect("Store is not initialized!") }
         }
     };
 }
